from dataclasses import dataclass
from typing import *


@dataclass
class Bbox:
    x1: int
    y1: int
    width: int
    height: int
    color: Optional[Tuple[int, int, int]] = None
    above_label: Optional[str] = None
    below_label: Optional[str] = None

    @classmethod
    def init_with_xyxy(cls, x1, y1, x2, y2, color=None, above_label=None, below_label=None):
        return cls(x1, y1, x2 - x1, y2 - y1, color, above_label, below_label)

    @property
    def x2(self):
        return self.x1 + self.width

    @property
    def y2(self):
        return self.y1 + self.height

    @property
    def x_center(self):
        return self.x1 + (self.width / 2)

    @property
    def y_center(self):
        return self.y1 + (self.height / 2)

    @property
    def area(self):
        return self.width * self.height



from typing import List

import numpy as np
import cv2

from . import BaseFrameEditCallback
from ..utils.video_player_utils import KeyFunction


class OpticalFlowPlotter(BaseFrameEditCallback):

    def __init__(
            self,
            enable_by_default: bool,
            min_arrow_size_to_draw: float = 10.0,
            draw_every_n_arrow: int = 80,
    ):
        super().__init__(enable_by_default)
        self._min_arrow_size_to_draw = min_arrow_size_to_draw
        self._draw_every_n_arrow = draw_every_n_arrow

    @property
    def enabled(self):
        return self._enabled

    def before_frame_resize(self, video_player, frame: np.ndarray, frame_num: int) -> np.ndarray:
        if frame_num == 0:
            return frame

        prev_frame = video_player.frame_reader.get_frame(frame_num - 1)
        gray_frame, prev_gray_frame = self._convert_frames_to_gray(frame, prev_frame)
        flow = self._calc_optical_flow(gray_frame, prev_gray_frame)
        frame = self._put_optical_flow_arrows_on_frame(frame, optical_flow_image=flow)

        return frame

    def enable_disable(self):
        self._enabled = not self._enabled

    @property
    def key_function_to_register(self) -> List[KeyFunction]:
        return [KeyFunction("o", func=self.enable_disable, description="show optical flow")]

    def _put_optical_flow_arrows_on_frame(self, frame, optical_flow_image):
        # Turn grayscale to rgb if needed
        if len(frame.shape) == 2:
            frame = np.stack((frame,) * 3, axis=2)

        # Get start and end coordinates of the optical flow
        flow_start = np.stack(np.meshgrid(range(optical_flow_image.shape[1]), range(optical_flow_image.shape[0])), 2)
        flow_end = (flow_start + optical_flow_image[flow_start[:, :, 1], flow_start[:, :, 0]] * 3).astype(np.int32)

        # Threshold values
        norm = np.linalg.norm(flow_end - flow_start, axis=2)
        norm = norm * (norm > self._min_arrow_size_to_draw)

        # Draw all the nonzero values
        nz = np.nonzero(norm)
        for i in range(0, len(nz[0]), self._draw_every_n_arrow):
            y, x = nz[0][i], nz[1][i]
            cv2.arrowedLine(
                frame,
                pt1=tuple(flow_start[y, x]),
                pt2=tuple(flow_end[y, x]),
                color=(0, 255, 0),
                thickness=1,
                tipLength=0.2,
            )

        return frame

    @staticmethod
    def _calc_optical_flow(gray_frame, prev_gray_frame) -> np.ndarray:
        flow = cv2.calcOpticalFlowFarneback(
            prev_gray_frame,
            gray_frame,
            flow=None,
            pyr_scale=0.5,
            levels=5,
            winsize=11,
            iterations=5,
            poly_n=5,
            poly_sigma=1.1,
            flags=0,
        )

        return flow

    @staticmethod
    def _convert_frames_to_gray(frame, prev_frame):
        if len(frame.shape) == 3 and frame.shape[2] == 3:
            gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            prev_gray_frame = cv2.cvtColor(prev_frame, cv2.COLOR_BGR2GRAY)
        else:
            gray_frame = frame
            prev_gray_frame = prev_frame
        return gray_frame, prev_gray_frame



import abc
import random
from pathlib import Path
import mimetypes
from typing import Optional

import cv2
import numpy as np

RANDOM_STATE = random.Random(42)


class FrameReader(abc.ABC):
    """
    The frame reader is used in the video player to fetch frames according to a frame number
    """

    @abc.abstractmethod
    def get_frame(self, frame_num: int) -> Optional[np.ndarray]:
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass


class LocalFrameReader(FrameReader):
    """
    A frame reader used to read any local video file or folder containing the frames as images
    (as long as there is a number in the name of the image files indicating their order).
    """

    def __init__(self, source_path):
        """
        Args:
            source_path: can be either a local video file path or a path to a directory containing frames as single
            images. the images need to contain a integer in their name specifying the frame order.
        """
        if Path(source_path).is_dir():
            self._reader = LocalDirReader(source_path)
        elif mimetypes.guess_type(source_path)[0].startswith("video"):
            self._reader = LocalVideoFileReader(source_path)
        else:
            raise NotImplementedError(f"{source_path=} is not a video file or directory")

    def get_frame(self, frame_num: int):
        return self._reader.get_frame(frame_num)

    def __len__(self):
        return len(self._reader)


class LocalVideoFileReader(FrameReader):
    def __init__(self, local_video_path: str):
        self._video_path = Path(local_video_path)
        assert self._video_path.is_file()
        self._video_reader = cv2.VideoCapture(str(self._video_path))
        self._total_frames = int(self._video_reader.get(cv2.CAP_PROP_FRAME_COUNT))

    def get_frame(self, frame_num):
        if frame_num >= len(self):
            return
        self._video_reader.set(cv2.CAP_PROP_POS_FRAMES, frame_num)
        ret, frame = self._video_reader.read()
        assert ret, f"frame number = {frame_num} is corrupted"
        return frame

    def __len__(self):
        return self._total_frames


class LocalDirReader(FrameReader):
    def __init__(self, local_frame_dir):
        self.local_frame_dir = local_frame_dir
        self._frame_paths = self._create_frame_list()

    def _create_frame_list(self):
        self._frame_paths = list(Path(self.local_frame_dir).glob("*"))
        self._frame_paths = [path for path in self._frame_paths if mimetypes.guess_type(path)[0].startswith("image")]

        if len(self._frame_paths) == 0:
            raise (Exception(f"No image files found in dir {self.local_frame_dir}"))

        self._frame_paths = sorted(self._frame_paths, key=lambda x: int(extract_digits_from_str(x.stem)))
        return self._frame_paths

    def get_frame(self, frame_num):
        if frame_num >= len(self):
            return
        img = cv2.imread(str(self._frame_paths[frame_num]), -1)
        assert img is not None, f"no image found in {self._frame_paths[frame_num]}"
        return img

    def __len__(self):
        return len(self._frame_paths)


def extract_digits_from_str(string):
    return "".join([char for char in string if char.isdigit()])



from typing import Tuple

import numpy as np

from ...utils.video_player_utils import KeyFunction
from ...utils.drawing_utils import write_text_on_img
from ..base_frame_edit_callback import BaseFrameEditCallback
from ...input_manager import InputManager


class KeyMapOverlay(BaseFrameEditCallback):
    def __init__(
        self,
        enable_by_default: bool = True,
        font_scale: float = 1,
        font_thickness: int = 1,
        font_color: Tuple[int, int, int] = (255, 255, 125),
        top_left_coordinate: Tuple[int, int] = (90, 10),
    ):
        super().__init__(enable_by_default)
        self._font_scale = font_scale
        self._font_thickness = font_thickness
        self._font_color = font_color
        self._tl_coordinate = top_left_coordinate

    @property
    def key_function_to_register(self):
        return [
            KeyFunction(key="ctrl+k", func=self.enable_disable, description="Show/Hide key map"),
        ]

    def after_frame_resize(self, video_player, frame: np.ndarray, frame_num: int) -> np.ndarray:
        row = self._tl_coordinate[0]
        write_text_on_img(
            frame,
            "Available keyboard shortcuts (hide with ctrl+k):",
            row=row,
            col=self._tl_coordinate[1],
            font_scale=self._font_scale * 1.2,
            thickness=self._font_thickness + 1,
            color=self._font_color,
        )
        row += int(30 * self._font_scale)
        for callback_name, keys in InputManager().get_keymap_description().items():
            write_text_on_img(
                frame,
                f"{callback_name}:",
                row=row,
                col=self._tl_coordinate[1],
                font_scale=self._font_scale,
                thickness=self._font_thickness + 1,
                color=self._font_color,
            )
            row += int(20 * self._font_scale)
            for key in keys:
                write_text_on_img(
                    frame,
                    key,
                    row=row,
                    col=self._tl_coordinate[1],
                    font_scale=self._font_scale,
                    thickness=self._font_thickness,
                    color=self._font_color,
                )
                row += int(20 * self._font_scale)
            row += int(10 * self._font_scale)
        return frame



import csv
from collections import defaultdict
from pathlib import Path
from typing import *

from cvvideoplayer.frame_editors import BaseBboxPlotter
from cvvideoplayer.utils.bbox_utils import Bbox


class DetectionsCsvPlotter(BaseBboxPlotter):
    """
    frame editor that plots detection given aa a csv is the following format:
    frame_id,label,x1,y1,width,height,score
    """

    def __init__(
        self,
        detections_csv_path: Path,
        enable_by_default: bool = True,
        show_class_label=True,
        **bbox_plotter_kwargs,
    ):
        super().__init__(enable_by_default, show_above_bbox_label=show_class_label, **bbox_plotter_kwargs)

        self._detections = defaultdict(list)
        with detections_csv_path.open(newline="") as csv_file:
            reader = csv.DictReader(csv_file)
            for row in reader:
                bbox = Bbox(
                    x1=int(float(row["x1"])),
                    y1=int(float(row["y1"])),
                    width=int(float(row["width"])),
                    height=int(float(row["height"])),
                    above_label=f"{row['label']} p: {float(row['score']):.2f}",
                )
                self._detections[int(row["frame_id"])].append(bbox)

    def get_bboxes(self, frame, frame_num) -> List[Bbox]:
        return self._detections[frame_num]



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple

import cv2
import numpy as np


class AbstractRecorder(ABC):

    @abstractmethod
    def write_frame_to_video(self, frame: np.ndarray) -> None:
        """
        method that receives a frame and writes in to a file
        """

    @abstractmethod
    def teardown(self) -> None:
        """
        what to do on closing of the player
        """


class SimpleRecorder(AbstractRecorder):
    """
    In charge of recording what ever the video player is playing. The output video will be saved in
    output_video_path
    """

    def __init__(
        self,
        output_video_path: Path = Path("./outputs/recorded_video.mp4"),
        recorded_video_fps: int = 30,
        output_video_shape: Tuple[int, int] = (640, 512),
    ):

        self._output_video_path = output_video_path
        self._recorded_video_fps = recorded_video_fps
        self._output_video_shape = output_video_shape
        self._video_writer = self._create_video_writer()

    def _create_video_writer(self):
        if self._output_video_path is None:
            return
        print(f"Saving video in {self._output_video_path}")
        self._output_video_path.parent.mkdir(exist_ok=True, parents=True)
        fourcc = cv2.VideoWriter_fourcc("m", "p", "4", "v")
        video_writer = cv2.VideoWriter(
            str(self._output_video_path),
            fourcc,
            self._recorded_video_fps,
            self._output_video_shape,
        )
        return video_writer

    def write_frame_to_video(self, frame):
        frame = cv2.resize(frame, self._output_video_shape)
        self._video_writer.write(frame)

    def teardown(self):
        self._video_writer.release()



import platform
import time
from pathlib import Path
from typing import Optional, List, Union
from functools import partial

import cv2
import numpy as np

from .input_manager import InputManager, SupportedOS
from .frame_editors import BaseFrameEditCallback
from .frame_reader import FrameReader
from .recorder import AbstractRecorder
from .utils.video_player_utils import (
    get_screen_adjusted_frame_size,
    KeyFunction,
    is_window_closed_by_mouse_click,
    get_screen_size,
    get_in_focus_window_id,
    get_frame_reader,
    get_recorder,
)


class VideoPlayer:
    def __init__(
        self,
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
    ):

        self.frame_reader = get_frame_reader(video_source)
        self._recorder = get_recorder(record)
        self._last_frame = len(self.frame_reader) - 1
        self._current_frame_num = start_from_frame - 1
        self._current_system = SupportedOS(platform.system())
        self._screen_size = get_screen_size(self._current_system)

        self._current_frame = None
        self._frame_edit_callbacks = frame_edit_callbacks

        self._play = False

        self._resize_factor = 1.0
        self._play_speed = 1
        self._exit = False
        self._window_name = "CVvideoPlayer"
        self._add_default_key_functions()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._recorder is not None:
            self._recorder.teardown()
        for callback in self._frame_edit_callbacks:
            callback.teardown()

    def run(self) -> None:
        self._setup()

        InputManager().start()

        while not self._exit:

            key = InputManager().get_input()

            if is_window_closed_by_mouse_click(window_name=self._window_name):
                cv2.destroyAllWindows()
                break
            if key == "mouse_click":
                cv2.waitKey(30)
            elif get_in_focus_window_id(self._current_system) != self._window_id:
                continue
            else:
                InputManager().handle_key_str(key)
                self._play_continuously() if self._play else self._show_current_frame()

        cv2.destroyAllWindows()

    def _setup_callbacks(self):
        for callback in self._frame_edit_callbacks:
            assert isinstance(callback, BaseFrameEditCallback), (
                "frame_editor must be a derived class of" " BaseFrameEditor"
            )
            for key_function in callback.key_function_to_register:
                assert isinstance(key_function, KeyFunction), (
                    f"{callback.__class__.__name__} is trying to register a key function"
                    f" (key = {key_function.key}) which is not an instance of KeyFunction"
                )
                InputManager().register_key_function(key_function, callback.__class__.__name__)

            callback.setup(self, self._current_frame)

    def _setup(self) -> None:
        self._change_current_frame(change_by=1)
        self._setup_callbacks()
        self._show_current_frame()
        time.sleep(0.5)  # make sure video player is up before checking the window id
        self._window_id = get_in_focus_window_id(self._current_system)
        cv2.waitKey(50)

    def _show_current_frame(self) -> None:
        frame = self._current_frame.copy()

        for callback in self._frame_edit_callbacks:
            if not callback.enabled:
                continue
            shape_before_edit = frame.shape[:2]
            frame = callback.before_frame_resize(self, frame, self._current_frame_num)
            assert frame.shape[:2] == shape_before_edit, "callbacks can not alter the frame's shape before resize"

        frame = self._resize_frame(frame)

        for callback in self._frame_edit_callbacks:
            if not callback.enabled:
                continue
            frame = callback.after_frame_resize(self, frame, self._current_frame_num)

        frame = self._resize_frame(frame)

        if self._recorder is not None:
            self._recorder.write_frame_to_video(frame)
        cv2.imshow(winname=self._window_name, mat=frame)
        cv2.waitKey(10)
        cv2.waitKey(1)  # for some reason Windows OS requires an additional waitKey to work properly

    def _play_continuously(self) -> None:
        while (not InputManager().has_input()) and self._play:
            self._change_current_frame(change_by=self._play_speed)
            self._show_current_frame()

    def _resize_frame(self, frame) -> np.ndarray:
        width, height = get_screen_adjusted_frame_size(
            screen_size=self._screen_size,
            frame_width=frame.shape[1],
            frame_height=frame.shape[0],
        )
        frame_size = int(self._resize_factor * width), int(self._resize_factor * height)
        return cv2.resize(frame, frame_size)

    def _change_current_frame(self, change_by: int) -> None:
        if change_by > 0 and self._current_frame_num == self._last_frame:
            self._play = False
            return

        if change_by < 0 and self._current_frame_num == 0:
            return

        self._current_frame_num = max(0, min(self._current_frame_num + change_by, self._last_frame))
        self._current_frame = self.frame_reader.get_frame(self._current_frame_num)

    def _pause_and_change_current_frame(self, change_by: int) -> None:
        self._play = False
        self._change_current_frame(change_by)

    def _change_frame_resize_factor(self, change_by: float) -> None:
        self._resize_factor = max(0.1, min(1.0, self._resize_factor + change_by))

    def _play_pause(self):
        self._play = not self._play

    def _increase_play_speed(self) -> None:
        if self._play_speed <= -1:
            self._play_speed = 1
        else:
            self._play_speed = min(16, self._play_speed * 2)

        self._play = True
        self._play_continuously()

    def _decrease_play_speed(self) -> None:
        if self._play_speed >= 1:
            self._play_speed = -1
        else:
            self._play_speed = max(-16, self._play_speed * 2)

        self._play = True
        self._play_continuously()

    def _on_exit(self) -> None:
        self._exit = True

    def _add_default_key_functions(self) -> None:
        default_key_functions = [
            KeyFunction("space", self._play_pause, "Play/Pause video"),
            KeyFunction("right", partial(self._pause_and_change_current_frame, 1), "Next frame"),
            KeyFunction("left", partial(self._pause_and_change_current_frame, -1), "Previous frame"),
            KeyFunction("ctrl+right", partial(self._pause_and_change_current_frame, 10), "10 frames forward"),
            KeyFunction("ctrl+left", partial(self._pause_and_change_current_frame, -10), "10 frames back"),
            KeyFunction("ctrl+shift+right", partial(self._pause_and_change_current_frame, 50), "50 frames forward"),
            KeyFunction("ctrl+shift+left", partial(self._pause_and_change_current_frame, -50), "50 frames back"),
            KeyFunction("+", partial(self._change_frame_resize_factor, 0.1), "Increase frame size"),
            KeyFunction("ctrl+=", partial(self._change_frame_resize_factor, 0.1), ""),
            KeyFunction("shift++", partial(self._change_frame_resize_factor, 0.1), ""),
            KeyFunction("=", partial(self._change_frame_resize_factor, 0.1), ""),
            KeyFunction("-", partial(self._change_frame_resize_factor, -0.1), "Decrease frame size"),
            KeyFunction("esc", self._on_exit, "Exit gracefully"),
        ]

        for key_function in default_key_functions:
            InputManager().register_key_function(key_function, "Video Control")



from typing import Tuple

import numpy as np

from ..base_frame_edit_callback import BaseFrameEditCallback
from ...utils.video_player_utils import KeyFunction
from ...utils.drawing_utils import write_text_on_img


class FrameInfoOverlay(BaseFrameEditCallback):
    def __init__(
        self,
        enable_by_default: bool = True,
        font_scale: float = 2,
        font_thickness: int = 2,
        top_left_coordinate: Tuple[int, int] = (10, 10),
    ):
        super().__init__(enable_by_default)
        self._video_total_frame_num = None
        self._font_scale = font_scale
        self._font_thickness = font_thickness
        self._tl_coordinate = top_left_coordinate
        self._orig_res = None

    @property
    def key_function_to_register(self):
        return [
            KeyFunction(key="ctrl+f", func=self.enable_disable, description="Show/Hide frame info"),
        ]

    def setup(self, video_player, frame) -> None:
        self._video_total_frame_num = len(video_player.frame_reader)
        self._orig_res = frame.shape

    def after_frame_resize(self, video_player, frame: np.ndarray, frame_num: int) -> np.ndarray:
        text = f"{frame_num}"
        if self._video_total_frame_num is not None:
            text += f"/{self._video_total_frame_num - 1}"

        line = write_text_on_img(
            frame,
            text,
            row=self._tl_coordinate[0],
            col=self._tl_coordinate[1],
            font_scale=self._font_scale,
            thickness=self._font_thickness,
        )

        write_text_on_img(
            frame,
            f"{self._orig_res[1]}x{self._orig_res[0]}",
            row=line,
            col=self._tl_coordinate[1],
            font_scale=self._font_scale / 1.5,
            thickness=self._font_thickness,
        )
        return frame



from typing import List, TYPE_CHECKING

import numpy as np

from ..utils.video_player_utils import KeyFunction

if TYPE_CHECKING:
    from .. import VideoPlayer


class BaseFrameEditCallback:

    def __init__(self, enable_by_default):
        self._enabled = enable_by_default

    @property
    def enabled(self):
        return self._enabled

    def setup(self, video_player: "VideoPlayer", frame) -> None:
        """
        Optionally configure more parameters according to the first incoming frame
        """

    def teardown(self) -> None:
        """
        Optionally define how the editor should close when video player is closed
        """

    def before_frame_resize(self, video_player: "VideoPlayer", frame: np.ndarray, frame_num: int) -> np.ndarray:
        """
        This function receives the frame before it has been resized and should return the frame
        after it has been altered in any way desirable by the user. In this hook you ere not allowed to change the
        frame's size.

        Args:
            video_player: an instance fo VideoPlayer
            frame (): the input frame
            frame_num ():

        Returns: the edited frame
        """
        return frame

    def after_frame_resize(self, video_player: "VideoPlayer", frame: np.ndarray, frame_num: int) -> np.ndarray:
        """
        This function receives the frame after it has been resized to fit the screen size and should return the frame
        after it has been altered in any way desirable by the user

        Args:
            video_player: an instance fo VideoPlayer
            frame (): the input frame
            frame_num ():

        Returns: the edited frame
        """
        return frame

    def enable_disable(self):
        self._enabled = not self._enabled

    @property
    def key_function_to_register(self) -> List[KeyFunction]:
        """
        Optionally return a list of KeyFunctions to be registered once the frame editor is added to the video player
        Examples:
            1. [
                  KeyFunction(
                      key="ctrl+f",  # A standard key combination can be any combination of modifiers with a single key
                      func=self.enable_disable,  # A function that receives no input
                      description="disable enable"  # A description of whatever this frame editor does
                  ),
                  KeyFunction(
                      key="ctrl+num", # A general num function that handles any number press
                      func=self._change_id, # this function needs to receive the number as an input
                      description="change the printed id according to the number pressed"
                  )
              ]
        """
        return []



from collections import defaultdict
from typing import Dict, Union
from queue import Queue
import platform

from pynput import keyboard, mouse

from .utils.video_player_utils import MODIFIERS, KeyFunction, SupportedOS
from .utils.windows_vk_dict import VK_CODE_MAP


def make_vk_code_mapper(os: SupportedOS):
    if os == SupportedOS.LINUX:
        return chr
    elif os == SupportedOS.WINDOWS:
        return lambda x: VK_CODE_MAP[x]
    else:
        raise NotImplementedError(f"{os=} not supported")


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class InputManager(metaclass=Singleton):
    def __init__(self):
        self._ui_queue = Queue()
        self._keymap: Dict[str, KeyFunction] = {}
        self._modifiers = set()
        self._current_system = SupportedOS(platform.system())
        self._map_vk_code = make_vk_code_mapper(self._current_system)
        self._listeners = []
        self._keymap_description: Dict[str, list] = defaultdict(list)

    def register_key_function(self, key_function: KeyFunction, callback_name: str) -> None:
        key = key_function.key
        if "+" in key:
            modifiers = sorted([modifier for modifier in key.split("+")[:-1] if modifier])
            key_without_modifiers = key.split("+")[-1] or "+"
            key = "+".join(modifiers + [key_without_modifiers])
        else:
            key_without_modifiers = key
            modifiers = []

        if key_without_modifiers.isnumeric():
            general_num_key = "+".join(modifiers + ["num"])
            if general_num_key in self._keymap:
                raise KeyError(f"{general_num_key} is registered: {self._keymap[general_num_key].description}")

        if key in self._keymap:
            raise KeyError(
                f"{key} is already registered (Registered action description: {self._keymap[key].description})"
            )

        self._keymap[key] = key_function
        if key_function.description:
            self._keymap_description[callback_name].append(f"{key:20.20}: {key_function.description:60.60}")

    def unregister_key_function(self, key: str) -> None:
        if key in self._keymap:
            del self._keymap[key]

    def has_input(self) -> bool:
        return not self._ui_queue.empty()

    def get_input(self) -> str:
        key = self._ui_queue.get()
        if key not in MODIFIERS:
            key = self._convert_key_to_str(key)
        return key

    def handle_key_str(self, key_str: str) -> None:
        key_without_modifiers = key_str.split("+")[-1]
        if key_without_modifiers.isnumeric():
            general_num_key = key_str.replace(key_without_modifiers, "num")
        else:
            general_num_key = None

        if general_num_key in self._keymap:
            self._keymap[general_num_key].func(key_without_modifiers)
        elif key_str in self._keymap:
            self._keymap[key_str].func()
        else:
            print(f"{key_str} is not registered in the keymap")

    def start(self) -> None:
        self._listeners.append(
            keyboard.Listener(
                on_press=self._add_key_press_to_queue,
                on_release=self._add_key_release_to_queue,
            )
        )
        self._listeners.append(mouse.Listener(on_click=self._add_mouse_click_to_queue))

        for listener in self._listeners:
            listener.start()

    def stop(self) -> None:
        for listener in self._listeners:
            listener.stop()
        self._listeners.clear()

    def _add_mouse_click_to_queue(self, *_) -> None:
        if not self.has_input():
            self._add_key_press_to_queue("mouse_click")

    def _add_key_press_to_queue(self, key: Union[keyboard.Key, str]) -> None:
        if key in MODIFIERS:
            key = str(key).replace("Key.", "").split("_", maxsplit=1)[0]
            if key not in self._modifiers:
                self._modifiers.add(key)

        elif not self.has_input():  # To avoid a situation of execution build up due to slow execution time
            self._ui_queue.put(key)

    def _add_key_release_to_queue(self, key: Union[keyboard.Key, str]) -> None:
        if key in MODIFIERS:
            key = str(key).replace("Key.", "").split("_", maxsplit=1)[0]
            if str(key) in self._modifiers:
                self._modifiers.remove(key)

    def get_keymap_description(self) -> dict:
        return self._keymap_description

    def print_keymap(self):
        print(self.get_keymap_description())

    def _convert_key_to_str(self, key: Union[keyboard.Key, keyboard.KeyCode, str]) -> str:
        if key == "mouse_click":
            return key

        if str(key) == "<65437>":  # work around for a bug in pynput model that does not convert 5 for some reason
            key = "5"

        if hasattr(key, "vk") and key.vk is not None:
            key = self._map_vk_code(key.vk).lower()
        else:
            key = str(key).replace("'", "").replace("Key.", "")

        key = "+".join(sorted(self._modifiers) + [key])

        return key
